"""Matching windows."""

import logging
import re
from collections.abc import Generator

from i3ipc import Con, Connection

from .layout_files import WindowMatchExpression

logger = logging.getLogger(__name__)


def find_windows_on_workspace(
    match_expression: WindowMatchExpression, workspace: Con
) -> Generator[Con]:
    """Find all windows on a workspace given a match expression.

    Parameters:
        match_expression: The match expression to use.
        workspace: A workspace tree object.
    """

    leaves = list(workspace.leaves())
    logger.debug(
        f'Looking for windows on the current workspace "{workspace.name}" with {len(leaves)} leaves'
    )
    for leaf in leaves:
        if leaf.type not in ["con", "floating_con"]:
            logger.debug(
                f"Skipping leaf with type {leaf.type}: {leaf.name} ({leaf.window_title})"
            )
            continue

        if is_window_matching(leaf, match_expression):
            logger.debug(f"Matching leaf found: {leaf.name} ({leaf.window_title})")
            yield leaf

    logger.debug("Finished window search")


def is_window_matching(con: Con, match_expression: WindowMatchExpression) -> bool:
    """Check if a window matches the given match expression.

    Parameters:
        con: The window to check.
        match_expression: The match expression to use.

    Returns:
        True if the window matches the expression, False otherwise.
    """

    if con.app_id is not None:
        # The window is Wayland native
        logger.debug(f"Checking Wayland con: {con.app_id} ({con.name})")
        wayland = match_expression.wayland
        if wayland is None:
            return False
        assert (
            wayland.app_id is not None or wayland.title is not None
        ), "At least one Wayland match expression must be provided, this should be enforced by the model."
        return (
            wayland.app_id is None or re.match(wayland.app_id, con.app_id) is not None
        ) and (wayland.title is None or re.match(wayland.title, con.name) is not None)
    elif con.window_class is not None or con.window_instance is not None:
        # The window runs under XWayland
        logger.debug(
            f"Checking XWayland con: {con.window_class},{con.window_instance} ({con.window_title})"
        )
        x11 = match_expression.x11
        if x11 is None:
            return False
        assert (
            x11.class_ is not None or x11.instance is not None or x11.title is not None
        ), "At least one X11 match expression must be provided, this should be enforced by the model."
        return (
            (x11.title is None or re.match(x11.title, con.window_title) is not None)
            and (
                x11.class_ is None or re.match(x11.class_, con.window_class) is not None
            )
            and (
                x11.instance is None
                or re.match(x11.instance, con.window_instance) is not None
            )
        )
    else:
        return False


def find_current_workspace(connection: Connection) -> Con | None:
    """Find the current workspace.

    Parameters:
        connection: A connection to Sway

    Returns:
        The tree node of the current workspace or `None` if it is not available.
    """

    tree = connection.get_tree()
    focused = tree.find_focused()
    if focused is None:
        return None
    workspace = focused.workspace()
    return workspace



"""Utilities related to the Sway connection."""

import logging

from i3ipc import CommandReply, Con, Connection

logger = logging.getLogger(__name__)


def run_command(connection: Connection, command: str) -> None:
    """Run a command and checks the reply.

    Arguments:
        connection: The Sway connection to use.
        command: The command to run.

    Raises:
        RuntimeError: If the command fails.
    """
    logger.debug(f"Running command '{command}'")
    replies = connection.command(command)
    check_replies(replies)


def run_command_on(con: Con, command: str) -> None:
    """Run a command on the given container and checks the reply.

    Arguments:
        con: The container to run the command on.
        command: The command to run.

    Raises:
        RuntimeError: If the command fails.
    """
    logger.debug(f"Running command '{command}' on container {con.name} ({con.id})")
    replies = con.command(command)
    check_replies(replies)


def check_replies(replies: list[CommandReply]):
    """Check a list of replies for errors.

    Arguments:
        replies: The replies to check.

    Raises:
        RuntimeError: If at least one reply indicates a failure.
    """
    for reply in replies:
        logger.debug(f"Command raw reply: {reply.ipc_data}")
        if not reply.success:
            raise RuntimeError(f"Command failed: {reply.error}")


def get_focused_workspace(connection: Connection) -> Con | None:
    """Get the currently focused workspace.

    Arguments:
        connection: The Sway connection.

    Returns:
        The con of the focused workspace, or None if no workspace is focused.
    """

    tree = connection.get_tree()
    focused = tree.find_focused()
    if focused is None:
        logger.warning("No focused con found.")
        return None
    if focused.type != "workspace":
        focused = focused.workspace()
    if focused is None:
        logger.warning("No focused workspace found.")
        return None
    return focused


def find_con_by_id(connection: Connection, con_id: int) -> Con:
    """Finds a containers with the given con_id.

    Arguments:
        connection: The Sway connection to use.
        con_id: The con_ids of the containers to find.

    Returns:
        A list of containers with the given con_ids. The order of the list matches
        the order of the con_ids.

    Raises:
        RuntimeError: If no container with the given con_id is not found in the tree.

    See also:
        [sway_out.connection.find_cons_by_id] for a version that finds multiple
        containers at once.
    """

    return find_cons_by_id(connection, con_id)[0]


def find_cons_by_id(connection: Connection, *con_ids: int) -> tuple[Con, ...]:
    """Finds all containers with the given con_ids.

    Each call to this function results in one IPC call. So passing multiple
    con_ids at once reduces the number of IPC calls while also ensuring that
    all con objects are consistent.

    Arguments:
        connection: The Sway connection to use.
        con_ids: The con_ids of the containers to find.

    Returns:
        A list of containers with the given con_ids. The order of the list matches
        the order of the con_ids.

    Raises:
        RuntimeError: If a container with a given con_id is not found in the tree.
    """

    result = find_cons_by_id_if_exists(connection, *con_ids)
    missing_ids = [con_id for con_id, con in zip(con_ids, result) if con is None]
    if missing_ids:
        raise RuntimeError(
            f"Container(s) with con_ids {', '.join(str(i) for i in missing_ids)} not found in tree"
        )
    assert None not in result
    return result


def find_cons_by_id_if_exists(
    connection: Connection, *con_ids: int
) -> tuple[Con | None, ...]:
    """Finds all containers with the given con_ids if they exist.
    This function is similar to `find_cons_by_id`, but it does not raise an error
    if a container with a given con_id is not found. Instead, it returns None for
    those containers.

    Arguments:
        connection: The Sway connection to use.
        con_ids: The con_ids of the containers to find.

    Returns:
        A list of containers with the given con_ids or `None` if the respective
        con_id does not match that of an existing container. The order of the
        list matches the order of the con_ids.
    """

    tree = connection.get_tree()
    return tuple(tree.find_by_id(con_id) for con_id in con_ids)



"""Funcntionality related to the layout of windows on a workspace."""

import itertools
import logging
from typing import Literal

from i3ipc import Con, Connection

from .connection import (
    find_con_by_id,
    find_cons_by_id,
    find_cons_by_id_if_exists,
    run_command_on,
)
from .layout_files import ApplicationLaunchConfig, ContainerConfig, WorkspaceLayout
from .utils import get_con_description, is_window

logger = logging.getLogger(__name__)

MARK = "@"
"""The layouting algorithm requires a mark to move windows around.

This should be something that is not in use otherwise.
"""

RESIZE_ATTEMPTS = 5
"""Number of attempts to resize a container to the expected size."""

RESIZE_TOLERANCE_PERCENT = 1
"""The tolerance in percent for resizing a container.

Because the parent container size might not always be exactly divisible by the
ratio of the children, we allow a small tolerance so that the layout can still be
applied.
"""


def dissolve_layout(connection: Connection, workspace_con: Con) -> None:
    """Dissolves the layout of the given workspace.

    After this function, all windows are direct children of the workspace.

    Parameters:
        connection: A connection to sway.
        workspace_con: The workspace container to dissolve the layout for.
    """

    def dissolve_child_layout(child_con: Con):
        if not child_con.nodes:
            return

        assert not is_window(child_con)

        for grandchild in child_con.nodes:
            dissolve_child_layout(grandchild)

        initial_child_con = find_con_by_id(connection, child_con.id)

        logger.debug(f"Dissolving layout of {get_con_description(initial_child_con)}")

        # Set the layout to horizontal to ensure that moving the children works
        # correctly.
        run_command_on(initial_child_con, "layout splith")

        for grandchild in initial_child_con.nodes:
            while True:
                (updated_grandchild_con, updated_child_con) = find_cons_by_id_if_exists(
                    connection, grandchild.id, child_con.id
                )
                if updated_grandchild_con is None:
                    raise RuntimeError(
                        f"Container {get_con_description(grandchild)} has disappeared"
                    )
                if updated_child_con is None:
                    logger.debug(
                        f"Layout {get_con_description(child_con)} was dissolved"
                    )
                    break
                elif updated_grandchild_con.id in {
                    con.id for con in updated_child_con.nodes
                }:
                    if len(updated_child_con.nodes) == 1:
                        # If the layout has only one child, there is a sway command
                        # to remove the layout.
                        run_command_on(grandchild, "split none")
                    else:
                        # If there are multiple children, we have to move the
                        # grandchild out of the child layout.
                        logger.debug(
                            f"Moving {get_con_description(updated_grandchild_con)} to move it "
                            + f"out of {get_con_description(updated_child_con)}"
                        )
                        run_command_on(updated_grandchild_con, "move left")
                else:
                    logger.debug(
                        f"Finished moving {get_con_description(updated_grandchild_con)} "
                        + f"out of {get_con_description(child_con)}"
                    )
                    break

        logger.debug(
            f"Dissolved layout of {get_con_description(child_con)} successfully"
        )

    logger.debug(
        f"Dissolving layout for workspace {get_con_description(workspace_con)}"
    )
    for child in workspace_con.nodes:
        dissolve_child_layout(child)

    # Sanity check
    for child in workspace_con.nodes:
        if child.nodes:
            logger.warning(
                f"There are still {len(child.nodes)} child nodes left on "
                + f"{get_con_description(child)} after dissolving the layout "
                + "on workspace {get_con_description"
            )
    logger.info(
        f"Dissolved layout for workspace {get_con_description(workspace_con)} successfully"
    )


def create_layout(
    connection: Connection,
    workspace_layout: WorkspaceLayout,
) -> None:
    """Creates a layout on the given workspace using the already existing windows.

    The algorithm is similar to insertion sort.

    All launch configurations have to have a con_id set.  The con_id attributes
    of containers will be set in the process.

    Parameters:
        connection: A connection to sway.
        workspace_layout: The layout to create.

    Note: This function modifies its argument.
    """

    def find_parent_con(con_id: int) -> Con:
        tree = connection.get_tree()
        for con in tree.descendants():
            if con_id in [c.id for c in con.nodes]:
                return con
        assert (
            False
        ), "This should not happen because there should always be at least a workspace as a parent."

    def move_con_to_workspace(con_id: int):
        (con,) = find_cons_by_id(connection, con_id)
        if con.workspace().id != workspace_id:
            logger.debug(
                f"Moving container {get_con_description(con)} to workspace {get_con_description(workspace_con)}"
            )
            run_command_on(
                con, f"move container to workspace {get_con_description(workspace_con)}"
            )
        else:
            logger.debug(
                f"Container {get_con_description(con)} is already on workspace {get_con_description(workspace_con)}"
            )

        # Make sure that the con is a direct child of the workspace to make layouting less error-prone.
        while find_parent_con(child_id).id != workspace_id:
            # Move the container to the right to not disturb the finished part of the layout.
            run_command_on(con, f"move right")
            # Make shure that the container is still on the workspace.
            (con,) = find_cons_by_id(connection, con_id)
            assert con.workspace().id == workspace_id, (
                f"Accidentally moved {get_con_description(con)} to another workspace "
                + f"({get_con_description(con.workspace())} "
                + f"instead of {get_con_description(workspace_con)})"
            )

    def swap_cons(con_id: int, target_id: int):
        (con, target_con) = find_cons_by_id(connection, con_id, target_id)
        if target_id != con_id:
            logger.debug(
                f"Swapping container {get_con_description(con)} with {get_con_description(target_con)} "
                + f"to position {index} on workspace {get_con_description(workspace_con)}"
            )
            run_command_on(con, f"swap container with con_id {target_con.id}")
        else:
            logger.debug(f"Container {get_con_description(con)} is already in position")

    def move_con_into(con_id: int, target_id: int):
        # There does not seem to be a way to move a con to an arbitrary position in a layout.
        # But we can move it into the layout using marks.
        (con, target_con) = find_cons_by_id(connection, con_id, target_id)
        run_command_on(target_con, f"mark --add {MARK}")
        run_command_on(con, f"move container to mark {MARK}")
        run_command_on(target_con, f"unmark {MARK}")

    def create_container_layout(
        container_layout: ApplicationLaunchConfig | ContainerConfig,
    ) -> int:
        if isinstance(container_layout, ContainerConfig):
            assert container_layout.children, "ContainerConfig must have children"
            logger.debug(
                f"Creating layout for container as {container_layout.layout} ..."
            )

            # First, find or create the first container in the layout.
            first_child_id = create_container_layout(container_layout.children[0])

            # Then create the layout.
            (first_child_con,) = find_cons_by_id(connection, first_child_id)
            run_command_on(first_child_con, f"splith")
            (first_child_con,) = find_cons_by_id(connection, first_child_id)
            run_command_on(first_child_con, f"layout {container_layout.layout}")
            layout_con = find_parent_con(first_child_con.id)
            layout_id = layout_con.id

            # Finally add the remaining children to the layout.
            for index, child_layout in itertools.islice(
                enumerate(container_layout.children), 1, None
            ):
                child_id = create_container_layout(child_layout)

                # Ensure that the child is on the right workspace.
                move_con_to_workspace(child_id)

                # Move the child into the layout.
                move_con_into(child_id, layout_id)
                assert find_parent_con(child_id).id == layout_id, (
                    f"The child {child_id} ended up somewhere unexpected after moving it into the "
                    + f"layout {layout_id}."
                )

                # Swap the child to the correct position if needed.
                (layout_con, child_con) = find_cons_by_id(
                    connection, layout_id, child_id
                )
                assert len(layout_con.nodes) >= index + 1, (
                    f"After moving the child, there should be at least {index + 1} windows on the layout, "
                    + "but found {len(layout_con.nodes)}"
                )
                swap_cons(child_id, layout_con.nodes[index].id)

            logger.debug(f"Layout for container as {container_layout.layout} done")
            layouted_ids.add(layout_id)
            container_layout._con_id = layout_id
            return layout_id
        else:
            # Nothing to lay out here.
            assert isinstance(container_layout, ApplicationLaunchConfig)
            tree = connection.get_tree()
            result = _find_con(tree, container_layout)
            logger.debug(f'Reached leaf "{result.name}" while creating layout')
            layouted_ids.add(result.id)
            return result.id

    # Check if the mark is already in use.
    marks_in_use = connection.get_marks()
    if MARK in marks_in_use:
        raise RuntimeError(f"The mark '{MARK}' is already in use.")

    # Track which cons we have layouted to detect windows that are not supposed to be here.
    layouted_ids = set()

    # Start the layout creation at the workspace level.
    # Set the layout of the workspace to horizontal to ensure moving containers to the workspace work correctly.
    workspace_id = workspace_layout._con_id
    assert workspace_id is not None, "The con_id should have been set earlier"
    (workspace_con,) = find_cons_by_id(connection, workspace_id)
    assert (
        workspace_con.nodes
    ), f"The workspace {get_con_description(workspace_con)} should not be empty at this point"
    run_command_on(workspace_con.nodes[0], "layout splith")
    for index, child_layout in enumerate(workspace_layout.children):
        logger.debug(
            f"Creating layout for workspace {get_con_description(workspace_con)} ..."
        )
        child_id = create_container_layout(child_layout)

        # Ensure that the child is on the right workspace.
        move_con_to_workspace(child_id)

        # Because we start at index == 0, there should now be at least index+1 windows on the workspace.
        (workspace_con, child_con) = find_cons_by_id(connection, workspace_id, child_id)
        assert len(workspace_con.nodes) >= index + 1

        # Swap the child to the correct position if needed.
        target_con = workspace_con.nodes[index]
        swap_cons(child_id, target_con.id)

    assert (
        workspace_con.nodes
    ), f"The workspace {get_con_description(workspace_con)} should not be empty at this point"
    if workspace_layout.layout is not None:
        # Set the layout of the workspace to the one specified in the layout.
        logger.debug(
            f"Setting layout of workspace {get_con_description(workspace_con)} to {workspace_layout.layout}"
        )
        run_command_on(workspace_con.nodes[0], f"layout {workspace_layout.layout}")
    logger.info(f"Layout for workspace {get_con_description(workspace_con)} done")
    layouted_ids.add(workspace_id)

    # The mark should be freed up after the layout is created.
    marks_in_use = connection.get_marks()
    assert (
        MARK not in marks_in_use
    ), f"The mark '{MARK}' was not removed after layout creation."


def find_leftover_windows(
    connection: Connection, workspace_layout: WorkspaceLayout
) -> list[Con]:
    """Finds windows that are not part of the given workspace layout.

    Parameters:
        connection: A connection to sway.
        workspace_layout: The layout to check against.

    Returns:
        A list of windows that are not part of the layout.
    """

    def remove_matched_windows(
        con_layout: WorkspaceLayout | ContainerConfig | ApplicationLaunchConfig,
    ):
        if isinstance(con_layout, ApplicationLaunchConfig):
            assert (
                con_layout._con_id is not None
            ), "The con_id of the application launch config should have been set before calling this function."
            if con_layout._con_id in leftover_windows:
                del leftover_windows[con_layout._con_id]
            else:
                logger.warning(
                    f"Application {con_layout.cmd} with con_id {con_layout._con_id} not found while "
                    + "looking for leftover windows."
                )
        else:
            for child in con_layout.children:
                remove_matched_windows(child)

    tree = connection.get_tree()
    workspace_id = workspace_layout._con_id
    leftover_windows = {
        con.id: con for con in tree.leaves() if con.workspace().id == workspace_id
    }
    remove_matched_windows(workspace_layout)
    return leftover_windows.values()


def resize_layout(
    connection: Connection,
    workspace_layout: WorkspaceLayout,
) -> None:
    """Resizes the containers on the workspace so that they match the given layout.


    Parameters:
        connection: A connection to sway.
        workspace_layout: The layout to apply.
    """

    def resize_children(con_layout: WorkspaceLayout | ContainerConfig):
        con = _find_con(tree, con_layout)

        con_width_px, con_height_px = get_container_size_excluding_gaps(con)

        if con_layout.layout in ["splith", "splitv"]:
            layout = con_layout.layout
        else:
            layout = "other"

        # Give up after a few attempts to avoid infinite loops.
        for i in range(RESIZE_ATTEMPTS):
            logger.debug(
                f"Resizing container {get_con_description(con)} to {con_width_px}px x {con_height_px}px "
                + f"(attempt {i + 1}/{RESIZE_ATTEMPTS})"
            )
            retry = False
            for child in con_layout.children:
                if not resize_con(child, con_width_px, con_height_px, layout):
                    retry = True
            if not retry:
                logger.debug(
                    f"Container {get_con_description(con)} resized successfully after "
                    + f"{i+1}/{RESIZE_ATTEMPTS} attempts."
                )
                break
        else:
            logger.error(
                f"Failed to resize container {get_con_description(con)} to {con_width_px}px x {con_height_px}px "
                + f"after {RESIZE_ATTEMPTS} attempts."
            )

    def resize_con(
        con_layout: ApplicationLaunchConfig | ContainerConfig,
        parent_width_px: int,
        parent_height_px: int,
        parent_layout: Literal["splith", "splitv", "other"],
    ) -> bool:
        con = _find_con(tree, con_layout)

        result = True
        if con_layout.percent is not None:
            assert parent_layout != "other", (
                "Percentages are only supported for horizontal and vertical splits, not for other layouts."
                + "This should have been enforced by the configuration models."
            )
            if parent_layout == "splith":
                expected_width_px = parent_width_px * con_layout.percent // 100
                run_command_on(con, f"resize set width {expected_width_px} px")
                actual_width_px = con.rect.width
                tolerance_px = expected_width_px * RESIZE_TOLERANCE_PERCENT // 100
                logger.debug(
                    f"Expected width: {expected_width_px}px, actual width: {actual_width_px}px, "
                    + f"tolerance: {tolerance_px}px"
                )
                result = (
                    expected_width_px - tolerance_px
                    <= actual_width_px
                    <= expected_width_px + tolerance_px
                )
            else:
                expected_height_px = parent_height_px * con_layout.percent // 100
                run_command_on(con, f"resize set height {expected_height_px} px")
                actual_height_px = con.rect.height
                tolerance_px = expected_height_px * RESIZE_TOLERANCE_PERCENT // 100
                logger.debug(
                    f"Expected height: {expected_height_px}px, actual height: {actual_height_px}px, "
                    + f"tolerance: {tolerance_px}px"
                )
                result = (
                    expected_height_px - tolerance_px
                    <= actual_height_px
                    <= expected_height_px + tolerance_px
                )

        # Resize children once the parent container has been resized.
        if result and isinstance(con_layout, ContainerConfig):
            resize_children(con_layout)

        return result

    tree = connection.get_tree()
    resize_children(workspace_layout)


def check_layout(
    connection: Connection,
    workspace_layout: WorkspaceLayout,
) -> bool:
    """Checks if the current layout matches the given workspace layout.

    Parameters:
        connection: A connection to sway.
        workspace_layout: The layout to check against.

    Returns:
        `True` if the current layout matches the given workspace layout, `False` otherwise.
    """

    def check_container_layout(
        container_layout: ApplicationLaunchConfig | ContainerConfig,
        parent_width_px: int,
        parent_height_px: int,
        parent_layout: Literal["splith", "splitv", "other"],
    ) -> bool:
        assert (
            container_layout._con_id is not None
        ), "The con_id of the container layout should have been set before calling this function."

        con = tree.find_by_id(container_layout._con_id)

        result = True

        if container_layout.percent is not None:
            assert parent_layout != "other", (
                "Percentages are only supported for horizontal and vertical splits, not for other layouts."
                + "This should have been enforced by the configuration models."
            )
            if parent_layout == "splith":
                # We assume that the decoration is always at the top.
                assert con.rect.width == con.deco_rect.width or con.deco_rect.width == 0
                actual_percent = con.rect.width * 100 // parent_width_px
                expected_width_px = parent_width_px * container_layout.percent // 100
                logger.debug(
                    f"Expected width: {expected_width_px}px, actual width: {con.rect.width}px"
                )
            else:
                actual_percent = (
                    (con.rect.height + con.deco_rect.height) * 100 // parent_height_px
                )
                expected_height_px = parent_height_px * container_layout.percent // 100
                logger.debug(
                    f"Expected height: {expected_height_px}px, actual height: {con.rect.height}px"
                )

            if (
                container_layout.percent - RESIZE_TOLERANCE_PERCENT
                <= actual_percent
                <= container_layout.percent + RESIZE_TOLERANCE_PERCENT
            ):
                logger.debug(
                    f"Container layout for {get_con_description(con)} matches the expected percentage: "
                    + f"{actual_percent}% == {container_layout.percent}%"
                )
                result = True
            else:
                logger.error(
                    f"Container layout for {get_con_description(con)} does not match the expected "
                    + f"percentage: {actual_percent}% != {container_layout.percent}%"
                )
                result = False

        if isinstance(container_layout, ContainerConfig):
            # Get the container dimensions
            con_width_px, con_height_px = get_container_size_excluding_gaps(con)
            if con.layout in ["splith", "splitv"]:
                con_layout = con.layout
            else:
                con_layout = "other"

            # Check all children, even if some do not match.
            for child in container_layout.children:
                if not check_container_layout(
                    child, con_width_px, con_height_px, con_layout
                ):
                    result = False

        return result

    tree = connection.get_tree()
    assert (
        workspace_layout._con_id is not None
    ), "The con_id of the workspace layout should have been set before calling this function."
    workspace_con = tree.find_by_id(workspace_layout._con_id)
    if not workspace_con:
        logger.error(
            f"Workspace with con_id {workspace_layout._con_id} not found in the tree."
        )
        return False

    if workspace_con.layout in ["splith", "splitv"]:
        layout = workspace_con.layout
    else:
        layout = "other"
    width_px, height_px = get_container_size_excluding_gaps(workspace_con)

    # Always check all containers in the workspace to show an exhaustive error message.
    result = True
    for child in workspace_layout.children:
        if not check_container_layout(child, width_px, height_px, layout):
            result = False
    return result


def get_container_size_excluding_gaps(con: Con) -> tuple[int, int]:
    # For windows, we use the rect and deco rect.
    # rect does not include the decoration, i.e. title bar.
    # For containers, we sum up the children where necessary to exclude
    # gaps between windows in the result.
    # We assume that the decoration is always at the top.

    if con.layout == "splith":
        width = sum(get_container_size_excluding_gaps(child)[0] for child in con.nodes)
    else:
        assert con.rect.width == con.deco_rect.width or con.deco_rect.width == 0
        width = con.rect.width

    if con.layout == "splitv":
        height = sum(get_container_size_excluding_gaps(child)[1] for child in con.nodes)
    else:
        height = con.rect.height + con.deco_rect.height

    return width, height


def _find_con(
    tree: Con, container: WorkspaceLayout | ContainerConfig | ApplicationLaunchConfig
) -> Con:
    con_id = container._con_id
    assert con_id is not None, (
        f"Application {container} has no con_id set. "
        + "This is required for the layout creation."
    )
    result = tree.find_by_id(con_id)
    if result is None:
        raise RuntimeError(f"Container for application with con ID {con_id} not found")
    else:
        return result



"""Small utility function."""

from i3ipc import Con

PROG_NAME = "sway-out"


def is_window(con: Con) -> bool:
    """Check if a container is a window.

    Parameters:
        con: The container to check.

    Returns:
        `True` if the container is a window, `False` otherwise.
    """
    return con.pid is not None


def get_con_description(con: Con) -> str:
    """Get a human-readable description of a container.

    Parameters:
        con: The container to describe.

    Returns:
        A string describing the container.
    """

    if is_window(con):
        return f"{con.name} (PID: {con.pid})"
    elif con.type == "workspace":
        return f"{con.name} (ID: {con.id})"
    else:
        return f"{con.name} (ID: {con.id})"



"""Launching of applications."""

import logging
import time

from i3ipc import Con, Connection

from .connection import check_replies
from .layout_files import (
    ApplicationLaunchConfig,
    ContainerConfig,
    WindowMatchExpression,
    WorkspaceLayout,
)
from .matching import find_current_workspace, find_windows_on_workspace
from .utils import get_con_description

logger = logging.getLogger(__name__)

LAUNCH_TIMEOUT_SECONDS = 10
"""How long to wait for the application to launch before giving up.

See also:
  - [wait_for_window][sway_out.applications.wait_for_window]
"""

LAUNCH_CHECK_INTERVAL_SECONDS = 0.5
"""How long to wait between checks for the application window.

See also:
  - [wait_for_window][sway_out.applications.wait_for_window]
"""


def match_existing_windows(workspace: Con, layout: WorkspaceLayout) -> None:
    """Match existing windows in the workspace with the launch configurations.

    This function updates the con_id of the launch configurations in the layout
    to match existing windows on the current workspace.

    Parameters:
        workspace: The workspace to search for existing windows.
        layout: The layout containing the applications to match.

    Note:
        This function modifies its argument.
    """

    def match_element(
        element_layout: ApplicationLaunchConfig | ContainerConfig,
    ) -> None:
        if isinstance(element_layout, ApplicationLaunchConfig):
            matching_windows = [
                con
                for con in find_windows_on_workspace(element_layout.match, workspace)
                if con.id not in matched_con_ids
            ]
            if matching_windows:
                con = matching_windows[0]
                element_layout._con_id = con.id
                matched_con_ids.add(con.id)
                logger.info(f"Matched existing window {get_con_description(con)}")
            else:
                logger.debug(f"No matching window found for {element_layout.cmd}")
        else:
            assert isinstance(element_layout, ContainerConfig)
            for child in element_layout.children:
                match_element(child)

    matched_con_ids: set[int] = set()

    for child in layout.children:
        match_element(child)

    logger.debug(f"Matched {len(matched_con_ids)} existing windows in the layout")


def launch_applications_from_layout(connection: Connection, layout: WorkspaceLayout):
    """Launch the applications contained in the given layout.

    The con_id of the launched applications are stored in
    the launch configurations for later reference.

    Parameters:
        connection: A connection to Sway.
        layout: The layout containing the applications to launch.

    Note:
        This function modifies its argument.

    See also:
        - [sway_out.applications.launch_application][]
    """

    def go(container: ApplicationLaunchConfig | ContainerConfig) -> None:
        if isinstance(container, ApplicationLaunchConfig):
            if hasattr(container, "_con_id"):
                logger.debug(
                    f"Skipping launch of {container.cmd} because it matched an existing window"
                )
            else:
                launch_application(connection, container)
        else:
            assert isinstance(container, ContainerConfig)
            for child in container.children:
                go(child)

    for child in layout.children:
        go(child)


def escape_argument(arg: str) -> str:
    """Escape a string for the use in a Sway command.

    Parameters:
        arg: The argument string.

    Returns:
        The argument string with some characters escaped.
    """
    # TODO: Implement
    return f'"{arg}"'


def launch_application(connection: Connection, launch_config: ApplicationLaunchConfig):
    """Launch an application on the current workspace.

    The con_id of the launched application is stored in the
    launch configuration for later reference.

    Parameters:
        connection: A connection to Sway.
        launch_config: The launch configuration for the application.

    Raises:
        RuntimeError:
            If the application fails to start.

    Note:
        This function modifies its argument.
    """

    if isinstance(launch_config.cmd, str):
        cmd = launch_config.cmd
    else:
        cmd = " ".join(escape_argument(a) for a in launch_config.cmd)

    # Find the currently focused workspace to launch the application on.
    workspace = find_current_workspace(connection)
    if workspace is None:
        logger.warning("No focused workspace found to search for windows.")
        raise RuntimeError("No focused workspace found to search for windows.")
    logger.debug(f"Preparing to launch command on focused workspace: {workspace.name}")

    # Find existing windows that match the launch configuration to ignore them.
    matching_windows_before = list(
        find_windows_on_workspace(launch_config.match, workspace)
    )
    logger.debug(
        f"{len(matching_windows_before)} window(s) match the expression before launch"
    )

    # Launch the application.
    logger.debug(f"Launching application with: '{cmd}'")
    replies = connection.command("exec " + cmd)
    check_replies(replies)

    # Wait for the application to launch and the window to appear
    try:
        con_id = wait_for_window(
            connection, workspace, launch_config.match, matching_windows_before
        )
    except RuntimeError as e:
        logger.error(f"Failed to launch application '{cmd}': {e}")
        raise RuntimeError(f"Failed to launch application '{cmd}': {e}") from e
    else:
        logger.info(f"'{cmd}' successfully launched with con_id {con_id}")
        launch_config._con_id = con_id


def wait_for_window(
    connection: Connection,
    workspace: Con,
    match: WindowMatchExpression,
    known_windows: list[Con],
) -> int:
    """Wait for the application to launch and a matching window to appear on the workspace.

    Parameters:
        connection: A connection to Sway.
        workspace: The workspace tree node to look in.
        match: The matching expression to use.
        known_windows: Windows to ignore.

    Returns:
        The Sway ID of the new con.

    Raises:
        RuntimeError:
            If no matching window is found within the timeout duration.
    """

    known_window_ids = {window.id for window in known_windows}
    matching_windows = []
    for _ in range(int(LAUNCH_TIMEOUT_SECONDS / LAUNCH_CHECK_INTERVAL_SECONDS)):
        # update the workspace to check for new windows
        workspace = connection.get_tree().find_by_id(workspace.id)
        if workspace is None:
            raise RuntimeError("The workspace has disappeared")
        logger.debug(f"Checking for matching windows")
        matching_windows = list(find_windows_on_workspace(match, workspace))
        logger.debug(
            f"{len(matching_windows)} window(s) currently match the launch expression {match}"
        )
        for window in matching_windows:
            if window.id not in known_window_ids:
                logger.debug(
                    f"New matching window found: {window.name} ({window.window_title})"
                )
                return window.id
        time.sleep(LAUNCH_CHECK_INTERVAL_SECONDS)
    else:
        raise RuntimeError(
            f"Application did not launch within {LAUNCH_TIMEOUT_SECONDS} seconds."
        )



"""Creation of new layout from existing workspace states."""

import logging

from i3ipc import Con, Connection

from sway_out.layout import get_container_size_excluding_gaps
from sway_out.layout_files import (
    ApplicationLaunchConfig,
    ContainerConfig,
    Layout,
    WaylandWindowMatchExpression,
    WindowMatchExpression,
    WorkspaceLayout,
    X11WindowMatchExpression,
)
from sway_out.utils import get_con_description, is_window

logger = logging.getLogger(__name__)


def create_layout_from_workspace(
    connection: Connection, workspace_names: list[str] | None = None
) -> Layout:
    """Create a layout from the current workspace states.

    Arguments:
        connection: A connection to Sway.
        workspace_names: The names of workspaces to include, iff omitted, all
            workspaces are included.

    Raises:
        RuntimeError: If an unexpected con type is encountered.

    Returns:
        A layout object.
    """

    def create_layout_for_container(con: Con, parent: Con):
        logger.debug("Creating layout for container %s", get_con_description(con))
        match con.type:
            case "con":
                if is_window(con):
                    if con.app_id is not None:
                        wayland = WaylandWindowMatchExpression(
                            app_id=con.app_id,
                            title=con.name,
                        )
                        x11 = None
                    else:
                        wayland = None
                        x11 = X11WindowMatchExpression(
                            class_=con.window_class,
                            instance=con.window_instance,
                            title=con.name,
                        )

                    result = ApplicationLaunchConfig(
                        cmd=_guess_command_for_application(con),
                        match=WindowMatchExpression(wayland=wayland, x11=x11),
                    )
                else:
                    children = [
                        create_layout_for_container(child, con) for child in con.nodes
                    ]
                    _fix_up_percentages(children)
                    result = ContainerConfig(
                        children=children,
                        layout=con.layout,
                    )

                if len(con.marks) == 1:
                    result.mark = con.marks[0]
                elif len(con.marks) > 1:
                    result.marks = con.marks

                if con.focused:
                    result.focus = True

                result.percent = _calculate_percent(con, parent)

                return result
            case "floating_con":
                logger.warning(
                    "Support for floating containers is not implemented, yet; skipping container %s",
                    get_con_description(con),
                )
            case _:
                raise RuntimeError(
                    f"Unexpected con type encountered for con_id {con.id}: {con.type}"
                )

    tree = connection.get_tree()
    workspaces: dict[str, WorkspaceLayout] = {}
    for workspace in tree.workspaces():
        if workspace_names is not None and workspace.name not in workspace_names:
            continue

        children = [
            create_layout_for_container(con, workspace) for con in workspace.nodes
        ]
        _fix_up_percentages(children)
        workspaces[workspace.name] = WorkspaceLayout(
            children=children, layout=workspace.layout
        )

    return Layout(focused_workspace=None, workspaces=workspaces)


def _guess_command_for_application(con: Con) -> list[str]:
    """Guess the command line for a container based on its PID.

    Parameters:
        con: The container to guess the command for.

    Returns:
        A list of command line arguments, or an empty list if the command cannot be guessed.
    """

    if con.pid is None:
        logger.error(
            "Cannot guess command for container %s without a PID",
            get_con_description(con),
        )
        return []

    # Retrieve the command line from /proc
    path = f"/proc/{con.pid}/cmdline"
    try:
        with open(path, "r") as f:
            cmdline = f.read().strip()
            if cmdline:
                # Join the command line arguments with spaces.
                # There seems to be a trailing null byte, so we drop the last element.
                return cmdline.split("\0")[:-1]
            else:
                return []
    except FileNotFoundError:
        logger.warning(f"Command line for PID {con.pid} not found at {path}")
        return []
    except Exception as e:
        logger.error(f"Error reading command line for PID {con.pid}: {e}")
        return []


def _calculate_percent(con: Con, parent: Con) -> int | None:
    """Calculate the percent for a container based on its size."""

    match parent.layout:
        case "stacked" | "tabbed":
            return None
        case "splith":
            # We assume that the decoration is always at the top.
            assert con.rect.width == con.deco_rect.width or con.deco_rect.width == 0
            child_width = con.rect.width
            parent_width = get_container_size_excluding_gaps(parent)[0]
            return child_width * 100 // parent_width
        case "splitv":
            child_height = con.rect.height + con.deco_rect.height
            parent_height = get_container_size_excluding_gaps(parent)[1]
            return child_height * 100 // parent_height
        case _:
            assert False or f"Unexpected layout type {parent.layout} encountered"

    if con.rect.width == 0 or con.rect.height == 0:
        return None
    return (con.rect.width * con.rect.height) / (
        con.workspace().rect.width * con.workspace().rect.height
    )


def _fix_up_percentages(
    children_layouts: list[ApplicationLaunchConfig | ContainerConfig],
):
    """Make the perentages total 100%.

    This is necessary in cases where the total is not exactly 100% due to rounding errors.

    Parameters:
        children_layouts: The list of children layouts to fix up.

    Note:
        This function modifies the input list in place.
    """

    # Skip the fix-up in cases where it is not necessary.
    if not children_layouts or any(child.percent is None for child in children_layouts):
        return

    total = sum(child.percent for child in children_layouts)
    difference = 100 - total
    if difference != 0:
        # Lazily adjust the last child to fix the percentage.
        last_child = children_layouts[-1]
        assert last_child.percent is not None, "This should have been checked before"
        children_layouts[-1].percent += difference

    assert (
        sum(child.percent for child in children_layouts) == 100
    ), "Percentages do not sum up to 100% after fix-up"



"""Data structures and utilities for layout descriptions."""

import re
from typing import Annotated, Literal, Self, TextIO

import yaml
from i3ipc import Connection
from pydantic import BaseModel, Field, PrivateAttr, field_validator, model_validator

from .connection import get_focused_workspace


class MarksMixin:
    """Mixin to add marks to a model.

    The fields related to marks are extracted to avoid duplication between
    [sway_out.layout_files.ApplicationLaunchConfig][] and
    [sway_out.layout_files.ContainerConfig][].
    """

    mark: Annotated[
        str | None,
        Field(
            default=None,
            title="Mark to assign",
            description="A mark to assign to the container. Mutually exclusive with `marks`.",
        ),
    ]
    marks: Annotated[
        list[str] | None,
        Field(
            default=None,
            title="Marks to assign",
            description="A list of marks to assign to the container. Mutually exclusive with `mark`.",
        ),
    ]

    @property
    def assigned_marks(self) -> list[str]:
        """Get the marks assigned to this container.
        Returns:
            A list of marks assigned to this container.
        """
        if self.mark is not None:
            return [self.mark]
        if self.marks is not None:
            return self.marks
        return []

    @field_validator("mark", "marks", mode="after")
    @staticmethod
    def validate_mark_values(v: str | list[str] | None) -> str | list[str] | None:
        if v is None:
            return v
        marks = v if isinstance(v, list) else [v]
        for mark in marks:
            if len(mark) != 1:
                raise ValueError(f"Mark must be a single character, got: {mark!r}")
        return v

    @model_validator(mode="after")
    def validate_marks(self) -> Self:
        if self.mark is not None and self.marks is not None:
            raise ValueError("Cannot set both `mark` and `marks` at the same time.")
        return self


class ConIdMixin:
    """Mixin to add con_id to a model.

    The con_id is used to identify the container in Sway.
    """

    _con_id: Annotated[int | None, PrivateAttr(default=None)]


class LayoutParentMixin:
    """A mixin to add layout-parent-related fields to a model."""

    children: "list[ApplicationLaunchConfig | ContainerConfig]"
    layout: Literal["splith", "splitv", "stacking", "tabbed"]

    @model_validator(mode="after")
    def validate_percent(self) -> Self:
        if self.layout in ["splith", "splitv"]:
            child_percentages = [
                child.percent or 0
                for child in self.children
                if child.percent is not None
            ]
            percent_sum = sum(child_percentages)
            any_none = (
                any(child.percent is None for child in self.children)
                or not self.children
            )
            if not any_none and percent_sum != 100:
                raise ValueError(
                    "If a percentage is set on all children of a"
                    + " layout, the percentages have add up to 100: "
                    + f"{' + '.join(str(p) for p in child_percentages)} = {percent_sum} != 100"
                )
        elif self.layout in ["stacking", "tabbed"]:
            any_percent = any(child.percent is not None for child in self.children)
            if any_percent:
                raise ValueError(
                    "Percentages are not allowed in stacking or tabbed layouts"
                )
        else:
            assert (
                False
            ), "Unknown layout type, this should have been caught by the validation."
        return self


class LayoutChildMixin:
    """A mixin to add layout-child-related fields to a model."""

    percent: Annotated[int | None, Field(default=None, ge=0, le=100)]


class FocusMixin:
    """A mixin to add focus-related fields to a model."""

    focus: Annotated[
        bool,
        Field(
            default=False,
            title="Focus after launch",
            description="If set, the container will be focused after it is launched. "
            + "Only one container is allowed to have set this to true.",
        ),
    ]


class WaylandWindowMatchExpression(BaseModel):
    app_id: str | None = None
    title: str | None = None

    @field_validator("app_id", "title", mode="after")
    @staticmethod
    def validate_regex(value: str | None) -> str | None:
        if value is None:
            return value
        try:
            _ = re.compile(value)
        except re.PatternError as e:
            raise ValueError(f"The regex '{value}' is invalid: {e}")
        else:
            return value

    @model_validator(mode="after")
    def validate_match(self) -> Self:
        if self.app_id is None and self.title is None:
            raise ValueError("At least one match expression must be provided.")
        return self


class X11WindowMatchExpression(BaseModel):
    class_: Annotated[str | None, Field(alias="class", title="Window class")] = None
    instance: Annotated[str | None, Field(title="Window instance")] = None
    title: str | None = None

    @field_validator("class_", "instance", "title", mode="after")
    @staticmethod
    def validate_regex(value: str | None) -> str | None:
        if value is None:
            return value
        try:
            _ = re.compile(value)
        except re.PatternError as e:
            raise ValueError(f"The regex '{value}' is invalid: {e}")
        else:
            return value

    @model_validator(mode="after")
    def validate_match(self) -> Self:
        if self.class_ is None and self.instance is None and self.title is None:
            raise ValueError("At least one match expression must be provided.")
        return self


class WindowMatchExpression(BaseModel):
    wayland: WaylandWindowMatchExpression | None = None
    x11: X11WindowMatchExpression | None = None

    @model_validator(mode="after")
    def validate_match(self) -> Self:
        if self.wayland is None and self.x11 is None:
            raise ValueError("At least one match expression must be provided.")
        return self


class ApplicationLaunchConfig(
    BaseModel, MarksMixin, ConIdMixin, LayoutChildMixin, FocusMixin
):
    cmd: Annotated[
        list[str] | str,
        Field(title="Launch command", description="Command to launch the application."),
    ]
    match: Annotated[
        WindowMatchExpression,
        Field(
            title="Window match expression",
            description="A filter to determine if the application is running.",
        ),
    ]


class ContainerConfig(
    BaseModel, MarksMixin, ConIdMixin, LayoutParentMixin, LayoutChildMixin, FocusMixin
):
    pass


class WorkspaceLayout(BaseModel, ConIdMixin, LayoutParentMixin):
    pass


class Layout(BaseModel):
    focused_workspace: Annotated[
        WorkspaceLayout | None,
        Field(
            default=None,
            title="Layout for the focused workspace",
            description="If present, this layout gets applied to the workspace that is currently focused.",
        ),
    ]
    workspaces: Annotated[
        dict[str, WorkspaceLayout] | None,
        Field(
            default=None,
            title="Workspace layouts",
            description="These layouts get applied to the workspaces with the name of the respective key.",
        ),
    ]

    @model_validator(mode="after")
    def validate_workspaces(self) -> Self:
        if self.focused_workspace is not None and self.workspaces is not None:
            raise ValueError(
                "Cannot set both `focused_workspace` and `workspaces` at the same time."
            )
        if self.focused_workspace is None and self.workspaces is None:
            raise ValueError(
                "Either `focused_workspace` or `workspaces` has to be present."
            )
        return self

    @model_validator(mode="after")
    def validate_focus(self) -> Self:
        def count_focused_elements(
            layout: WorkspaceLayout | ContainerConfig | ApplicationLaunchConfig,
        ) -> int:
            count = 0

            if isinstance(layout, (WorkspaceLayout, ContainerConfig)):
                count += sum(count_focused_elements(child) for child in layout.children)

            if (
                isinstance(layout, (ContainerConfig, ApplicationLaunchConfig))
                and layout.focus
            ):
                count += 1

            return count

        total_count = 0

        if self.focused_workspace is not None:
            total_count += count_focused_elements(self.focused_workspace)

        if self.workspaces is not None:
            for workspace_layout in self.workspaces.values():
                total_count += count_focused_elements(workspace_layout)

        if total_count > 1:
            raise ValueError(
                "Only one container can have `focus` set to `True` in the entire layout."
            )

        return self


def load_layout_configuration(file: TextIO) -> Layout:
    """Load a layout configuration from a file-like object.

    Arguments:
        file: The source file.

    Raises:
        yaml.YAMLError:
            When the file does not contain valid YAML.
        pydantic.ValidationError:
            When the YAML is ill-formed.

    Returns:
        The configuration object.
    """

    obj = yaml.load(file, yaml.SafeLoader)
    return Layout.model_validate(obj)


def save_layout_configuration(layout: Layout, file: TextIO):
    """Save a layout configuration to a file-like object.

    Arguments:
        layout: The layout to save.
        file: The destination file.
    """

    obj = layout.model_dump(exclude_none=True, exclude_unset=True)
    yaml.dump(obj, file, yaml.SafeDumper, encoding="utf-8", allow_unicode=True)


def map_workspaces(
    connection: Connection, layout: Layout
) -> dict[str, WorkspaceLayout]:
    """Map the workspace names from the layout to the actual workspaces.

    The currently focused workspace is used to resolve
    [sway_out.layout_files.Layout.focused_workspace][].

    con_ids stay unset if the workspace does not exist in Sway.

    Arguments:
        connection: A connection to Sway.
        layout: The layout to map.

    Returns:
        A mapping of workspace layouts with the con_id set to the id of the
        corresponding workspace to the name of the respective workspace.

    Note:
        The layout objects do not get copied, so `layout` is modified.
    """

    def go():
        for workspace_name, workspace_layout in layout.workspaces.items():
            for workspace in tree.workspaces():
                if workspace.name == workspace_name:
                    workspace_layout._con_id = workspace.id
                    break
            yield workspace_name, workspace_layout
        if layout.focused_workspace is not None:
            focused_workspace_layout = layout.focused_workspace
            focused_worksapce_con = get_focused_workspace(connection)
            assert focused_worksapce_con is not None, "No focused workspace found?"
            focused_workspace_name = focused_worksapce_con.name
            yield focused_workspace_name, focused_workspace_layout

    tree = connection.get_tree()
    return dict(go())


def find_focused_element_in_layout(
    layout: Layout,
) -> ApplicationLaunchConfig | ContainerConfig | None:
    """Find element to focus in a layout.

    Arguments:
        layout: The layout to search in.

    Returns:
        The focused element or `None` if no element is focused.
    """

    if layout.focused_workspace is not None:
        return find_focused_element_on_workspace(layout.focused_workspace)
    if layout.workspaces is not None:
        for workspace_layout in layout.workspaces.values():
            focused_element = find_focused_element_on_workspace(workspace_layout)
            if focused_element is not None:
                return focused_element


def find_focused_element_on_workspace(
    layout: WorkspaceLayout | ContainerConfig | ApplicationLaunchConfig,
) -> ApplicationLaunchConfig | ContainerConfig | None:
    """Find the element to focus in a layout.

    Arguments:
        layout: The layout to search in.

    Returns:
        The focused element or `None` if no element is focused.
    """

    if isinstance(layout, (ContainerConfig, ApplicationLaunchConfig)) and layout.focus:
        return layout
    if isinstance(layout, (WorkspaceLayout, ContainerConfig)):
        for child in layout.children:
            focused_child = find_focused_element_on_workspace(child)
            if focused_child is not None:
                return focused_child
    return None



"""Utilities to indicate progress."""

import logging
import subprocess
from contextlib import contextmanager
from typing import Generator, Literal, final

from sway_out.utils import PROG_NAME

logger = logging.getLogger(__name__)


@final
class ProgressNotification:
    """Represents a notification showing progress."""

    EXPIRE_TIME = 60_000  # 60 seconds

    def __init__(self, stage: str, text: str):
        self.summary = stage
        self.text = text
        self.successful = True
        self.notification_id: int | None = None

    def start(self):
        """Start the progress indicator.

        Without calling this method first, `self.update()` is a no-op.
        """

        self.notification_id = _run_notify_send(
            summary=self.summary,
            text=f"{self.text} ...",
            expire_time=self.EXPIRE_TIME,
        )

    def update(self, progress: int, total: int):
        """Update the progress.

        Does nothing if `self.start` is not called first.
        """

        if self.notification_id is not None:
            self.notification_id = _run_notify_send(
                summary=self.summary,
                text=f"{self.text} {progress}/{total}",
                expire_time=self.EXPIRE_TIME,
                replace_id=self.notification_id,
            )

    def finish(self):
        """Show that the process has finished.

        Does nothing if `self.start` is not called first.
        """

        if self.notification_id is not None:
            # We are intentionally not explicitly passing an expire_time to
            # respect use the user's configuration.
            if self.successful:
                self.notification_id = _run_notify_send(
                    summary=self.summary,
                    text="Completed successfully.",
                    replace_id=self.notification_id,
                )
            else:
                self.notification_id = _run_notify_send(
                    summary=self.summary,
                    text="Completed with errors.",
                    replace_id=self.notification_id,
                )

    def error(self, error_message: str):
        """Show that an error has occurred.

        Does nothing if `self.start` is not called first.
        """

        if self.notification_id is not None:
            self.notification_id = _run_notify_send(
                summary=self.summary,
                text=(f"{self.text} - " if self.text else "")
                + f"Error: {error_message}",
                urgency="critical",
                replace_id=self.notification_id,
            )


@contextmanager
def progress_notification(stage: str, text: str) -> Generator[ProgressNotification]:
    """Show a notification with progress updates.

    This function is intended to be used as a context manager.

    Parameters:
        stage: The stage of the process.
        text: The text to display in the notification.

    Returns:
        A callback function that can be used to update the progress.
    """

    notification = ProgressNotification(stage, text)
    try:
        yield notification
    except Exception as e:
        notification.error(f"{text} - Error: {e}")
        raise
    else:
        notification.finish()


def error_notification(title: str, text: str) -> None:
    """Show a notification indicating an error.

    Parameters:
        title: The title of the notification.
        text: The error message.
    """

    _run_notify_send(summary=title, text=text, urgency="critical")


def _run_notify_send(
    summary: str,
    text: str,
    urgency: Literal["low", "normal", "critical"] = "normal",
    replace_id: int | None = None,
    expire_time: int | None = None,
):
    command = [
        "notify-send",
        f"--app-name={PROG_NAME}",
        f"--urgency={urgency}",
        "--print-id",
        summary,
        text,
    ]
    if replace_id is not None:
        command.append(f"--replace-id={replace_id}")
    if expire_time is not None:
        command.append(f"--expire-time={expire_time}")
    result = subprocess.run(command, capture_output=True, text=True)
    result.check_returncode()
    notification_id = int(result.stdout.strip())
    logger.debug(f"Showing notification with ID {notification_id}: {summary} - {text}")
    return notification_id



"""Main entrypoint."""

import logging
from dataclasses import dataclass

import click
import pydantic
import yaml
from i3ipc import Connection

from sway_out.layout_creation import create_layout_from_workspace

from .applications import launch_applications_from_layout, match_existing_windows
from .connection import find_con_by_id, run_command, run_command_on
from .layout import (
    check_layout,
    create_layout,
    dissolve_layout,
    find_leftover_windows,
    resize_layout,
)
from .layout_files import (
    find_focused_element_in_layout,
    load_layout_configuration,
    map_workspaces,
    save_layout_configuration,
)
from .marks import apply_marks
from .matching import find_current_workspace
from .notifications import error_notification, progress_notification
from .utils import PROG_NAME, get_con_description

logger = logging.getLogger(__name__)


@dataclass
class GlobalState:
    """The user-provided configuration for the application and some global state."""

    connection: Connection
    notifications: bool = True


@click.group()
@click.version_option(prog_name=PROG_NAME, package_name="sway-out")
@click.option(
    "--notifications/--no-notifications",
    default=True,
    help="Enable or disable notifications.",
)
@click.pass_context
def main(ctx: click.Context, notifications: bool):
    """Main entrypoint."""

    logging.basicConfig(level=logging.DEBUG)
    ctx.obj = GlobalState(Connection(), notifications)


@main.command("apply")
@click.argument("layout_file", type=click.File("r"))
@click.pass_context
def main_apply(ctx: click.Context, layout_file):
    connection: Connection = ctx.obj.connection
    assert connection is not None
    try:
        configuration = load_layout_configuration(layout_file)
    except (yaml.YAMLError, pydantic.ValidationError) as e:
        if ctx.obj.notifications:
            error_notification("Error during layout creation", str(e))

        click.echo(f"Failed to read layout configuration: {e}", err=True)
        return

    workspace_layout_mapping = map_workspaces(connection, configuration)

    with progress_notification("Applying layout", "Workspace") as notification:
        if ctx.obj.notifications:
            notification.start()
        for index, (workspace_name, workspace_layout) in enumerate(
            workspace_layout_mapping.items()
        ):
            notification.update(index + 1, len(workspace_layout_mapping))
            run_command(connection, f"workspace {workspace_name}")
            workspace_con = find_current_workspace(connection)
            workspace_layout._con_id = workspace_con.id
            assert workspace_con is not None, "No current workspace found?"
            logging.debug("Matching existing windows on workspace: %s", workspace_name)
            match_existing_windows(workspace_con, workspace_layout)
            logging.debug("Dissolving layout for workspace: %s", workspace_name)
            dissolve_layout(connection, workspace_con)
            logger.info("Applying layout for workspace: %s", workspace_name)
            launch_applications_from_layout(connection, workspace_layout)
            create_layout(connection, workspace_layout)
            leftover_windows = find_leftover_windows(connection, workspace_layout)
            if not leftover_windows:
                resize_layout(connection, workspace_layout)
            else:
                error_message = (
                    f"Found leftover windows on workspace {workspace_name}:\n"
                    + "\n".join(f"- {get_con_description(w)}" for w in leftover_windows)
                    + "\n"
                    + "Not resizing layout."
                )
                click.echo(error_message, err=True)
                if ctx.obj.notifications:
                    error_notification("Applying layout", error_message)
                notification.successful = False
            apply_marks(connection, workspace_layout)
            if not leftover_windows and not check_layout(connection, workspace_layout):
                error_message = f"Failed to apply layout to {workspace_name}"
                click.echo(error_message, err=True)
                if ctx.obj.notifications:
                    error_notification("Applying layout", error_message)
                notification.successful = False

        focused_layout = find_focused_element_in_layout(configuration)
        if focused_layout is not None:
            focused_con = find_con_by_id(connection, focused_layout._con_id)
            run_command_on(focused_con, "focus")
            logger.info(
                f"Focused element in layout: {get_con_description(focused_con)}",
            )
        else:
            logger.debug("No focused element found in layout")

        logger.info(f"Applied layout for {len(workspace_layout_mapping)} workspace(s)")


@main.command("save")
@click.argument("layout_file", type=click.File("w"), required=False)
@click.option(
    "-w",
    "--workspace",
    type=str,
    required=False,
    multiple=True,
    help="Restrict to specific workspaces",
)
@click.pass_context
def main_save(ctx: click.Context, layout_file, workspace):
    connection: Connection = ctx.obj.connection
    assert connection is not None
    workspace_names = set(workspace) if workspace else None

    with progress_notification("Creating layout", "Creation") as notification:
        if ctx.obj.notifications:
            notification.start()
        layout = create_layout_from_workspace(connection, list(workspace) or None)
        save_layout_configuration(layout, layout_file)

    logger.info("Layout creation completed.")


if __name__ == "__main__":
    main()



"""Functions for managing marks on containers."""

from i3ipc import Connection

from .connection import find_cons_by_id, run_command_on
from .layout_files import ApplicationLaunchConfig, ContainerConfig, WorkspaceLayout


def has_marks(connection: Connection, workspace_layout: WorkspaceLayout) -> bool:
    """Check if the the workspace has all marks from the layout assigned.

    Parameters:
        connection: A connection to Sway.
        workspace_layout: The layout containing the marks to check for.

    Returns:
        True if the workspace has all marks assigned, False otherwise.
    """

    def go(layout: ContainerConfig | ApplicationLaunchConfig):
        (con,) = find_cons_by_id(connection, layout._con_id)

        return not (set(layout.assigned_marks) - set(con.marks)) and (
            not isinstance(layout, ContainerConfig)
            or all(go(child) for child in layout.children)
        )

    return all(go(child) for child in workspace_layout.children)


def apply_marks(connection: Connection, workspace_layout: WorkspaceLayout) -> None:
    """Apply marks to containers in the specified workspace layout.

    All nodes with marks in the workspace layout have to have their con_id set.

    Parameters:
        connection: A connection to Sway.
        workspace_layout: The layout containing the containers to mark.
    """

    def go(layout: ContainerConfig | ApplicationLaunchConfig):
        (con,) = find_cons_by_id(connection, layout._con_id)
        marks = layout.assigned_marks
        for m in marks:
            run_command_on(con, f"mark --add {m}")

        if isinstance(layout, ContainerConfig):
            for child in layout.children:
                go(child)

    for child in workspace_layout.children:
        go(child)



from . import (
    applications,
    connection,
    layout,
    layout_creation,
    layout_files,
    main,
    marks,
    matching,
    notifications,
    utils,
)

__all__ = [
    "applications",
    "connection",
    "layout",
    "layout_files",
    "layout_creation",
    "main",
    "marks",
    "matching",
    "notifications",
    "utils",
]
